from enum import IntEnum


class ThumbALUOpcode(IntEnum):
    AND = 0x0
    EOR = 0x1
    LSL = 0x2
    LSR = 0x3
    ASR = 0x4
    ADC = 0x5
    SBC = 0x6
    ROR = 0x7
    TST = 0x8
    NEG = 0x9
    CMP = 0xA
    CMN = 0xB
    ORR = 0xC
    MUL = 0xD
    BIC = 0xE
    MVN = 0xF
