from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.constants cimport CPUMode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport (
    get_bit,
    get_bits,
    set_bit,
    add_32,
)

cdef void arm_block_data_transfer(CPU, uint32_t) noexcept
